@@ -14,9 +14,8 @@ static PyMethodDef hiredis_ReaderMethods[] = {
 };
 
 PyTypeObject hiredis_ReaderType = {
-    PyObject_HEAD_INIT(NULL)
-    0,                            /*ob_size*/
-    "hiredis.Reader",             /*tp_name*/
+    PyVarObject_HEAD_INIT(NULL, 0)
+    MOD_HIREDIS ".Reader",        /*tp_name*/
     sizeof(hiredis_ReaderObject), /*tp_basicsize*/
     0,                            /*tp_itemsize*/
     (destructor)Reader_dealloc,   /*tp_dealloc*/
@@ -34,7 +33,7 @@ PyTypeObject hiredis_ReaderType = {
     0,                            /*tp_getattro*/
     0,                            /*tp_setattro*/
     0,                            /*tp_as_buffer*/
-    Py_TPFLAGS_DEFAULT,           /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
     "Hiredis protocol reader",    /*tp_doc */
     0,                            /*tp_traverse */
     0,                            /*tp_clear */
@@ -59,7 +58,7 @@ static void *tryParentize(const redisReadTask *task, PyObject *obj) {
     PyObject *parent;
     if (task && task->parent) {
         parent = (PyObject*)task->parent->obj;
-        assert(parent->ob_type == PyList_Type);
+        assert(PyList_CheckExact(parent));
         PyList_SET_ITEM(parent, task->idx, obj);
     }
     return obj;
@@ -69,13 +68,13 @@ static PyObject *createDecodedString(hiredis_ReaderObject *self, const char *str
     PyObject *obj;
 
     if (self->encoding == NULL) {
-        obj = PyString_FromStringAndSize(str, len);
+        obj = PyBytes_FromStringAndSize(str, len);
     } else {
         obj = PyUnicode_Decode(str, len, self->encoding, NULL);
         if (obj == NULL) {
             if (PyErr_ExceptionMatches(PyExc_ValueError)) {
                 /* Ignore encoding and simply return plain string. */
-                obj = PyString_FromStringAndSize(str, len);
+                obj = PyBytes_FromStringAndSize(str, len);
             } else {
                 assert(PyErr_ExceptionMatches(PyExc_LookupError));
 
@@ -98,16 +97,23 @@ static PyObject *createDecodedString(hiredis_ReaderObject *self, const char *str
     return obj;
 }
 
+static void *createError(PyObject *errorCallable, char *errstr, size_t len) {
+    PyObject *obj;
+
+    PyObject *args = Py_BuildValue("(s#)", errstr, len);
+    assert(args != NULL); /* TODO: properly handle OOM etc */
+    obj = PyObject_CallObject(errorCallable, args);
+    assert(obj != NULL);
+    Py_DECREF(args);
+    return obj;
+}
+
 static void *createStringObject(const redisReadTask *task, char *str, size_t len) {
     hiredis_ReaderObject *self = (hiredis_ReaderObject*)task->privdata;
     PyObject *obj;
 
     if (task->type == REDIS_REPLY_ERROR) {
-        PyObject *args = Py_BuildValue("(s#)", str, len);
-        assert(args != NULL); /* TODO: properly handle OOM etc */
-        obj = PyObject_CallObject(self->replyErrorClass, args);
-        assert(obj != NULL);
-        Py_DECREF(args);
+        obj = createError(self->replyErrorClass, str, len);
     } else {
         obj = createDecodedString(self, str, len);
     }
@@ -150,19 +156,15 @@ static void Reader_dealloc(hiredis_ReaderObject *self) {
     if (self->encoding)
         free(self->encoding);
 
-    self->ob_type->tp_free((PyObject*)self);
+    ((PyObject *)self)->ob_type->tp_free((PyObject*)self);
 }
 
 static int _Reader_set_exception(PyObject **target, PyObject *value) {
-    int subclass;
-    subclass = PyObject_IsSubclass(value, PyExc_Exception);
+    int callable;
+    callable = PyCallable_Check(value);
 
-    if (subclass == -1) {
-        return 0;
-    }
-
-    if (subclass == 0) {
-        PyErr_SetString(PyExc_TypeError, "Expected subclass of Exception");
+    if (callable == 0) {
+        PyErr_SetString(PyExc_TypeError, "Expected a callable");
         return 0;
     }
 
@@ -191,18 +193,24 @@ static int Reader_init(hiredis_ReaderObject *self, PyObject *args, PyObject *kwd
             return -1;
 
     if (encodingObj) {
+        PyObject *encbytes;
         char *encstr;
         int enclen;
 
-        encstr = PyString_AsString(encodingObj);
-        if (encstr != NULL) {
-            enclen = strlen(encstr);
-            self->encoding = (char*)malloc(enclen+1);
-            memcpy(self->encoding, encstr, enclen);
-            self->encoding[enclen] = '\0';
-        } else {
+        if (PyUnicode_Check(encodingObj))
+            encbytes = PyUnicode_AsASCIIString(encodingObj);
+        else
+            encbytes = PyObject_Bytes(encodingObj);
+
+        if (encbytes == NULL)
             return -1;
-        }
+
+        enclen = PyBytes_Size(encbytes);
+        encstr = PyBytes_AsString(encbytes);
+        self->encoding = (char*)malloc(enclen+1);
+        memcpy(self->encoding, encstr, enclen);
+        self->encoding[enclen] = '\0';
+        Py_DECREF(encbytes);
     }
 
     return 0;
@@ -213,12 +221,12 @@ static PyObject *Reader_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     self = (hiredis_ReaderObject*)type->tp_alloc(type, 0);
     if (self != NULL) {
         self->reader = redisReplyReaderCreate();
-        redisReplyReaderSetReplyObjectFunctions(self->reader, &hiredis_ObjectFunctions);
-        redisReplyReaderSetPrivdata(self->reader, self);
+        self->reader->fn = &hiredis_ObjectFunctions;
+        self->reader->privdata = self;
 
         self->encoding = NULL;
-        self->protocolErrorClass = HiErr_ProtocolError;
-        self->replyErrorClass = HiErr_ReplyError;
+        self->protocolErrorClass = HIREDIS_STATE->HiErr_ProtocolError;
+        self->replyErrorClass = HIREDIS_STATE->HiErr_ReplyError;
         Py_INCREF(self->protocolErrorClass);
         Py_INCREF(self->replyErrorClass);
 
@@ -242,11 +250,17 @@ static PyObject *Reader_feed(hiredis_ReaderObject *self, PyObject *args) {
 
 static PyObject *Reader_gets(hiredis_ReaderObject *self) {
     PyObject *obj;
-    char *err;
+    PyObject *err;
+    char *errstr;
 
     if (redisReplyReaderGetReply(self->reader, (void**)&obj) == REDIS_ERR) {
-        err = redisReplyReaderGetError(self->reader);
-        PyErr_SetString(self->protocolErrorClass, err);
+        errstr = redisReplyReaderGetError(self->reader);
+        /* protocolErrorClass might be a callable. call it, then use it's type */
+        err = createError(self->protocolErrorClass, errstr, strlen(errstr));
+        obj = PyObject_Type(err);
+        PyErr_SetString(obj, errstr);
+        Py_DECREF(obj);
+        Py_DECREF(err);
         return NULL;
     }
 
